from functools import wraps
from typing import Any, Literal

from iso4217 import Currency

from .client import RampAsyncClient
from .constants import (
    AMOUNT_DESCRIPTION,
    SK_CATEGORIES,
)
from .memory_db import MemoryDatabase
from .types import Role
from .utils import get_nested_keys, str_date_to_datetime

ramp_client = RampAsyncClient()
memory_db = MemoryDatabase()


def handle_response(func):
    """
    Wrap every tool with this to handle errors
    """

    @wraps(func)
    async def wrapper(*args, **kwargs) -> str | list[dict[str, Any]]:
        try:
            return await func(*args, **kwargs)
        except Exception as e:
            return (
                f"Ran into an error: {e}\n"
                "Communicate this to the user and consider retrying "
                "if the error seems transient.\n"
            )

    return wrapper


def handle_load_response(func):
    """
    Wrap every load tool with this
    """

    @wraps(func)
    @handle_response
    async def wrapper(*args, **kwargs) -> str:
        data = await func(*args, **kwargs)
        if not data:
            return "No data found\n"
        # hack to help us get past claude's context window and input size limitations
        name = memory_db.store_data(func.__name__, data)
        return (
            f"Stored data in memory database with table name: {name}.\n "
            f"Call `process_data` tool with table name and desired columns to setup a SQL table.\n "
            f"Call `execute_query` tool with query to get results as a JSON.\n "
            f"Available columns are: {', '.join(get_nested_keys(data[0]))}\n"
            f"Call `clear_table` tool with table name to delete the table from the memory database.\n"
        )

    return wrapper


# ================================================
# Database tools
# ================================================
@handle_response
async def clear_table(table_name: str) -> str:
    """
    Clear a table
    """
    memory_db.clear_table(table_name)
    return f"Table {table_name} cleared"


@handle_response
async def process_data(table_name: str, cols: list[str]) -> str:
    """
    Process data into a table
    """
    if memory_db.data_is_processed(table_name):
        return f"Table {table_name} already processed"
    memory_db.create_table_with_cols(table_name, cols)
    memory_db.load_data(table_name)
    memory_db.commit(table_name)
    return f"Table {table_name} created"


@handle_response
async def execute_query(table_name: str, query: str) -> list[dict[str, Any]] | str:
    f"""
    Execute a SQL query on a table and return results as a list of objects
    Use sqlite syntax and use window functions whenever possible.
    {AMOUNT_DESCRIPTION}
    """
    if not memory_db.exists(table_name):
        return f"Error: table {table_name} does not exist"
    return memory_db.execute_query(query)


# ================================================
# Fetch tools
# ================================================
@handle_response
async def get_ramp_categories() -> list[dict[str, Any]]:
    """
    Get Ramp transaction categories
    """
    return [{"id": k, "name": v} for k, v in SK_CATEGORIES.items()]


@handle_response
async def get_currencies() -> list[dict[str, Any]]:
    """
    Get currencies
    """
    return [
        {"currency_code": val.code, "currency_name": val.currency_name}
        for val in Currency
    ]


# ================================================
# Load tools
# ================================================
@handle_load_response
async def load_transactions(
    from_date: str,
    to_date: str,
    user_id: str | None = None,
    card_id: str | None = None,
    ramp_category_ids: list[str] = [],
    accounting_sync_ready: bool | None = None,
) -> list[dict[str, Any]]:
    f"""
    Transactions are sorted by amount in descending order and only represent card transactions.
    {AMOUNT_DESCRIPTION}
    """
    params = {
        "from_date": str_date_to_datetime(from_date),
        "to_date": str_date_to_datetime(to_date, add_one_day=True),
        "order_by_amount_desc": True,
    }
    if user_id:
        params["user_id"] = user_id
    if ramp_category_ids:
        params["sk_category_ids"] = ramp_category_ids
    if card_id:
        params["card_id"] = card_id
    if accounting_sync_ready is not None:
        params["sync_ready"] = accounting_sync_ready

    return await ramp_client.paginate_list_endpoint(
        "/transactions",
        params,
    )


@handle_load_response
async def load_spend_export(
    from_date: str,
    to_date: str,
) -> list[dict[str, Any]]:
    """
    Spend export is a list of all spend events (transactions, reimbursements, bills etc.) for a user.
    Always use this over load_transactions, load_reimbursements, load_bills, etc. when possible
    """
    params = {
        "from_date": str_date_to_datetime(from_date),
        "to_date": str_date_to_datetime(to_date, add_one_day=True),
    }
    if from_date:
        params["from_date"] = str_date_to_datetime(from_date)
    if to_date:
        params["to_date"] = str_date_to_datetime(to_date, add_one_day=True)

    return await ramp_client.paginate_list_endpoint(
        "/spend-export",
        params,
    )


@handle_load_response
async def load_receipts(
    from_date: str,
    to_date: str,
    transaction_id: str | None = None,
    created_before: str | None = None,
    created_after: str | None = None,
) -> list[dict[str, Any]]:
    f"""
    Get receipts
    {AMOUNT_DESCRIPTION}
    """
    params = {
        "from_date": str_date_to_datetime(from_date),
        "to_date": str_date_to_datetime(to_date, add_one_day=True),
        "order_by_amount_desc": True,
    }
    if transaction_id:
        params["transaction_id"] = transaction_id
    if created_before:
        params["created_before"] = str_date_to_datetime(created_before)
    if created_after:
        params["created_after"] = str_date_to_datetime(created_after, add_one_day=True)

    return await ramp_client.paginate_list_endpoint(
        "/receipts",
        params,
    )


@handle_load_response
async def load_reimbursements(
    from_date: str,
    to_date: str,
    sync_ready: bool | None = None,
    direction: Literal["BUSINESS_TO_USER", "USER_TO_BUSINESS", ""] = "",
    user_id: str | None = None,
) -> list[dict[str, Any]]:
    f"""
    Get reimbursements for user
    {AMOUNT_DESCRIPTION}
    """
    params = {
        "from_date": str_date_to_datetime(from_date),
        "to_date": str_date_to_datetime(to_date, add_one_day=True),
    }
    if sync_ready is not None:
        params["sync_ready"] = sync_ready
    if direction:
        params["direction"] = direction
    if user_id:
        params["user_id"] = user_id

    return await ramp_client.paginate_list_endpoint(
        "/reimbursements",
        params,
    )


@handle_load_response
async def load_bills(
    from_date: str,
    to_date: str,
    payment_status: Literal["OPEN", "PAID", ""] = "",
    user_id: str | None = None,
    acccounting_sync_ready: bool | None = None,
) -> list[dict[str, Any]]:
    f"""
    Get bills for user
    {AMOUNT_DESCRIPTION}
    """
    params = {
        "from_created_at": str_date_to_datetime(from_date),
        "to_created_at": str_date_to_datetime(to_date, add_one_day=True),
    }
    if user_id:
        params["user_id"] = user_id
    if acccounting_sync_ready is not None:
        params["sync_ready"] = acccounting_sync_ready
    if payment_status:
        params["payment_status"] = payment_status

    return await ramp_client.paginate_list_endpoint(
        "/bills",
        params,
    )


@handle_load_response
async def load_locations(
    entity_id: str | None = None,
) -> list[dict[str, Any]]:
    """
    Get locations for entity
    """
    params = {}
    if entity_id:
        params["entity_id"] = entity_id

    return await ramp_client.paginate_list_endpoint(
        "/locations",
        params,
    )


@handle_load_response
async def load_departments() -> list[dict[str, Any]]:
    """
    Get departments
    """
    return await ramp_client.paginate_list_endpoint(
        "/departments",
        {},
    )


@handle_load_response
async def load_bank_accounts(
    bank_account_id: str,
) -> list[dict[str, Any]]:
    """
    Get bank accounts for own business/entities.
    """
    params = {}
    if bank_account_id:
        params["bank_account_id"] = bank_account_id

    return await ramp_client.paginate_list_endpoint(
        "/bank_accounts",
        params,
    )


@handle_load_response
async def load_vendors(
    ramp_category_ids: list[str] = [],
    is_active: bool | None = True,
    name: str | None = None,
    from_created_at: str | None = None,
    to_created_at: str | None = None,
) -> list[dict[str, Any]]:
    """
    Get vendors for user

    Usually search for only active vendors.
    """
    params = {}
    if ramp_category_ids:
        params["sk_category_ids"] = ramp_category_ids
    if is_active is not None:
        params["is_active"] = is_active
    if name:
        params["name"] = name
    if from_created_at:
        params["from_created_at"] = str_date_to_datetime(from_created_at)
    if to_created_at:
        params["to_created_at"] = str_date_to_datetime(to_created_at, add_one_day=True)

    return await ramp_client.paginate_list_endpoint(
        "/vendors",
        params,
    )


@handle_load_response
async def load_vendor_bank_accounts(
    vendor_id: str,
) -> list[dict[str, Any]]:
    """
    Get bank accounts for vendors
    """
    params = {}

    return await ramp_client.paginate_list_endpoint(
        f"/vendors/{vendor_id}/accounts",
        params,
    )


@handle_load_response
async def load_entities(
    entity_name: str | None = None,
) -> list[dict[str, Any]]:
    """
    Get entities for user
    """
    params = {}
    if entity_name:
        params["entity_name"] = entity_name

    return await ramp_client.paginate_list_endpoint(
        "/entities",
        params,
    )


@handle_load_response
async def load_spend_limits(
    user_id: str | None = None,
) -> list[dict[str, Any]]:
    """
    Get limits for user
    """
    params = {}
    if user_id:
        params["user_id"] = user_id

    return await ramp_client.paginate_list_endpoint(
        "/limits",
        params,
    )


@handle_load_response
async def load_spend_programs() -> list[dict[str, Any]]:
    """
    Get spend programs
    """
    return await ramp_client.paginate_list_endpoint(
        "/spend-programs",
        {},
    )


@handle_load_response
async def load_users(
    email: str | None = None,
    role: Role = "",
) -> list[dict[str, Any]]:
    """
    Get users from Ramp
    """
    params = {}
    if email:
        params["email"] = email
    if role:
        params["role"] = role

    return await ramp_client.paginate_list_endpoint(
        "/users",
        params,
    )



CLIENT_MAX_PAGES = 100
AMOUNT_DESCRIPTION = (
    "The amount is an integer in smallest denomination to avoid precision loss.\n"
    "- So 1000 refers to 1000 cents or $10.00\n"
)
SK_CATEGORIES = {
    1: "Pet",
    2: "Other",
    3: "Office",
    4: "Airlines",
    5: "Car rental",
    6: "Lodging",
    7: "Travel misc",
    8: "Taxi and rideshare",
    9: "Freight, moving and delivery services",
    10: "Shipping",
    11: "Utilities",
    12: "Office supplies and cleaning",
    13: "General merchandise",
    14: "Electronics",
    15: "Clothing",
    16: "Books and newspapers",
    17: "Supermarkets and grocery stores",
    18: "Fuel and gas",
    19: "Restaurants",
    20: "Alcohol and bars",
    21: "Medical",
    23: "Fees and financial institutions",
    24: "Entertainment",
    25: "Professional services",
    26: "Taxes and tax preparation",
    27: "Advertising",
    28: "Parking",
    29: "Car services",
    30: "Gambling",
    31: "Clubs and memberships",
    32: "Legal",
    33: "Education",
    34: "Charitable donations",
    35: "Political organizations",
    36: "Religious organizations",
    37: "Fines",
    38: "Government services",
    39: "Intra-company purchases",
    40: "SaaS / Software",
    41: "Cloud computing",
    42: "Streaming services",
    43: "Internet and phone",
    44: "Insurance",
}



from typing import Literal, TypeAlias

# these should be enums
# claude likes to pass in empty strings to mean null
Interval: TypeAlias = Literal[
    "ANNUAL",
    "WEEKLY",
    "MONTHLY",
    "TOTAL",
    "",
]
Icon: TypeAlias = Literal[
    "AccountingServicesIcon",
    "AdvertisingIcon",
    "CONTRACTORS_AND_PROFESSIONAL_SERVICES",
    "CUSTOM",
    "CardIcon",
    "EducationStipendIcon",
    "EmployeeRewardsIcon",
    "GroundTransportationIcon",
    "LegalFeesIcon",
    "LodgingIcon",
    "LunchOrderingIcon",
    "OnboardingIcon",
    "PerDiemCardIcon",
    "SOFTWARE",
    "SaasSubscriptionIcon",
    "SoftwareTrialIcon",
    "SuppliesIcon",
    "TeamSocialIcon",
    "TravelExpensesIcon",
    "VirtualEventIcon",
    "WellnessIcon",
    "WorkFromHomeIcon",
    "",
]
Role: TypeAlias = Literal[
    "IT_ADMIN",
    "BUSINESS_ADMIN",
    "BUSINESS_OWNER",
    "BUSINESS_USER",
    "GUEST_USER",
    "",
]



import base64
import os
from typing import Any

import httpx

from .constants import CLIENT_MAX_PAGES

ENV_TO_BASE_URL = {
    "demo": "https://demo-api.ramp.com/developer/v1",
    "prd": "https://api.ramp.com/developer/v1",
}


def get_access_token_with_client_credentials(base_url: str, scopes: list[str]) -> str:
    client_id, client_secret = (
        os.getenv("RAMP_CLIENT_ID"),
        os.getenv("RAMP_CLIENT_SECRET"),
    )
    headers = {
        "Accept": "application/json",
        "Authorization": f"Basic {base64.b64encode(f'{client_id}:{client_secret}'.encode()).decode()}",
        "Content-Type": "application/x-www-form-urlencoded",
    }
    payload = {
        "grant_type": "client_credentials",
        "scope": " ".join(scopes),
    }

    response = httpx.post(
        f"{base_url}/token",
        headers=headers,
        data=payload,
    )
    response.raise_for_status()

    return response.json()["access_token"]


class RampAsyncClient:
    """
    Lightweight wrapper around Ramp's Developer API
    """

    def connect(self, scopes: list[str]):
        if not self._access_token:
            self._access_token = get_access_token_with_client_credentials(
                self._base_url,
                scopes,
            )
        self.client = httpx.AsyncClient(
            headers={
                "Accept": "application/json",
                "Authorization": f"Bearer {self._access_token}",
                "Content-Type": "application/json",
                "User-Agent": "Ramp-MCP/0.0.1",
            }
        )

    def __init__(self):
        env = os.getenv("RAMP_ENV", "demo")
        if env not in ENV_TO_BASE_URL:
            raise ValueError(f"Invalid Ramp environment: {env}")
        self._base_url = ENV_TO_BASE_URL[env]
        self._access_token = None
        self.client = None

    async def paginate_list_endpoint(
        self,
        path: str,
        params: dict[str, Any],
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """
        GET all pages of a list endpoint
        """
        results = []
        _url = f"{self._base_url}{path}"
        i = 1
        while _url is not None:
            if i > CLIENT_MAX_PAGES:
                raise Exception("Too many pages, try to filter more results out.")
            response = await self.client.get(
                _url,
                params=params | kwargs if not results else None,
            )
            response.raise_for_status()
            res = response.json()
            results.extend(res["data"])

            _url = res["page"]["next"]
            i += 1
        return results



import click
from mcp.server.fastmcp import FastMCP

from ramp_mcp.tools import (
    clear_table,
    execute_query,
    get_currencies,
    get_ramp_categories,
    load_bank_accounts,
    load_bills,
    load_departments,
    load_entities,
    load_locations,
    load_receipts,
    load_reimbursements,
    load_spend_export,
    load_spend_limits,
    load_spend_programs,
    load_transactions,
    load_users,
    load_vendor_bank_accounts,
    load_vendors,
    process_data,
    ramp_client,
)

server = FastMCP(
    "ramp-mcp",
    instructions="""
    Ramp is a platform for corporate cards, bill payments, accounting,
    and a whole lot more. Data is made available via the developer API.
    - For performance, always try to load all of the data needed
    prior to setting up tables or running queries. Use as many window functions
    as possible.
    - Ensure that all calculations are accurate.
    """,
    dependencies=["httpx", "iso4217", "flatten_json"],
)

scope_to_tools_mapping = {
    # load tools
    "transactions:read": [load_transactions],
    "receipts:read": [load_receipts],
    "reimbursements:read": [load_reimbursements],
    "bills:read": [load_bills],
    "locations:read": [load_locations],
    "departments:read": [load_departments],
    "bank_accounts:read": [load_bank_accounts],
    "vendors:read": [load_vendors, load_vendor_bank_accounts],
    "entities:read": [load_entities],
    "limits:read": [load_spend_limits],
    "spend_programs:read": [load_spend_programs],
    "users:read": [load_users],
}


@click.command()
@click.option(
    "--scopes",
    "-s",
    type=click.STRING,
    default="transactions:read,reimbursements:read,bills:read",
    help=f"""
    Comma seperated list of scopes to use. Do not include spaces.
    Available scopes: {", ".join(scope_to_tools_mapping.keys())}
    """,
)
def main(scopes: str):
    # register load tools based on scopes
    scopes = scopes.split(",")
    for scope in scopes:
        if scope not in scope_to_tools_mapping:
            continue
        for tool in scope_to_tools_mapping[scope]:
            server.add_tool(tool)

    # special tool that requires all three scopes
    if all(
        scope in scopes
        for scope in ["transactions:read", "reimbursements:read", "bills:read"]
    ):
        server.add_tool(load_spend_export)

    # always register fetch tools
    server.add_tool(get_ramp_categories)
    server.add_tool(get_currencies)
    server.add_tool(clear_table)
    server.add_tool(process_data)
    server.add_tool(execute_query)

    ramp_client.connect(scopes)

    server.run(transport="stdio")


if __name__ == "__main__":
    main()


__all__ = ["main", "server"]



import datetime
from typing import Any


def str_date_to_datetime(date_str: str, add_one_day: bool = False) -> datetime.datetime:
    date = datetime.datetime.strptime(date_str, "%Y-%m-%d")
    if add_one_day:
        # easier than explaining exclusive end date
        date += datetime.timedelta(days=1)
    return date.astimezone(datetime.timezone.utc)


def get_nested_keys(data: dict[str, Any]) -> list[str]:
    """
    Get all keys including nested keys as a dot-separated string
    e.g. {"a": {"b": "c"}, "d": "e"} -> ["a__b", "d"]
    """
    stack = [(key, value, key) for key, value in data.items()]
    keys = []

    while stack:
        _, value, path = stack.pop()

        if isinstance(value, dict):
            for sub_key, sub_value in value.items():
                if path == sub_key:
                    # hack for tie breakers
                    new_path = f"{path}__{sub_key}_1"
                else:
                    new_path = f"{path}__{sub_key}"
                stack.append((sub_key, sub_value, new_path))
        else:
            keys.append(path)

    return keys



import datetime
import json
import sqlite3
import uuid
from typing import Any

from flatten_json import flatten


class MemoryDatabase:
    """
    ephemeral sqlite db in memory to make json array data queryable by an llm
    with simple built-in ETL data pipeline
    """

    def __init__(self):
        self.conn = sqlite3.connect(":memory:")
        self.conn.row_factory = sqlite3.Row
        self._cursor = self.conn.cursor()

        self._table_name_to_temp_data: dict[str, list[dict[str, Any]]] = dict()
        self._table_name_to_cols: dict[str, list[str]] = dict()
        self._table_last_access: dict[str, datetime.datetime] = dict()

    def store_data(self, func_name: str, data: list[dict[str, Any]]) -> str:
        """
        Store response data for processing
        """
        table_name = func_name + "_" + str(uuid.uuid4()).replace("-", "")
        self._table_name_to_temp_data[table_name] = data
        return table_name

    def data_is_processed(self, table_name: str) -> bool:
        """
        Check if a table has been processed
        """
        return table_name not in self._table_name_to_temp_data

    def exists(self, table_name: str) -> bool:
        """
        Check if a table exists
        """
        return table_name in self._table_last_access

    def commit(self, table_name: str) -> None:
        """
        Commit the current transaction
        """
        self.conn.commit()
        self._table_last_access[table_name] = datetime.datetime.now()
        if table_name in self._table_name_to_temp_data:
            del self._table_name_to_temp_data[table_name]

    def create_table_with_cols(
        self,
        table_name: str,
        cols: list[str],
    ) -> None:
        """
        Create a table from a list of columns
        """
        processed_data = self._process_data(
            self._table_name_to_temp_data[table_name], cols
        )
        self._table_name_to_temp_data[table_name] = processed_data

        unique_cols = {key for data in processed_data for key in data.keys()}
        self._table_name_to_cols[table_name] = list(unique_cols)

        column_types = self._infer_column_types(unique_cols, processed_data)

        columns = []
        for key, col_type in column_types.items():
            columns.append(f'"{key}" {col_type}')

        create_table_sql = (
            f'CREATE TABLE IF NOT EXISTS "{table_name}" ({", ".join(columns)})'
        )
        self._cursor.execute(create_table_sql)

    def load_data(self, table_name: str) -> None:
        """
        Insert data into table from temp data store
        """
        data = self._table_name_to_temp_data[table_name]
        if not data:
            return

        unique_cols = self._table_name_to_cols[table_name]
        placeholders = ", ".join(["?" for _ in unique_cols])
        columns_str = ", ".join(unique_cols)

        insert_sql = f"INSERT INTO {table_name} ({columns_str}) VALUES ({placeholders})"

        values = [tuple(item.get(col, None) for col in unique_cols) for item in data]
        self._cursor.executemany(insert_sql, values)

    def execute_query(self, query: str) -> list[dict[str, Any]]:
        """
        Execute a SQL query and return results as a list of dicts
        """
        try:
            self._cursor.execute(query)
            rows = self._cursor.fetchall()
            return [dict(row) for row in rows]
        except sqlite3.Error as e:
            raise Exception(f"SQL error: {e}")

    def clear_table(self, table_name: str) -> None:
        """
        Clear a table
        """
        self._cursor.execute(f"DROP TABLE IF EXISTS {table_name}")
        if table_name in self._table_name_to_temp_data:
            del self._table_name_to_temp_data[table_name]
        if table_name in self._table_name_to_cols:
            del self._table_name_to_cols[table_name]
        if table_name in self._table_last_access:
            del self._table_last_access[table_name]
        self.conn.commit()

    def _infer_column_types(
        self, cols: list[str], data: list[dict[str, Any]]
    ) -> dict[str, str]:
        """
        Infer column types from data
        """
        column_types = {col: "INTEGER" for col in cols}
        for data in data:
            for key, value in data.items():
                if isinstance(value, str) or value is None:
                    column_types[key] = "TEXT"
                elif isinstance(value, float):
                    column_types[key] = "REAL"

        return column_types

    def _get_value_from_key(
        self,
        data: dict[str, Any],
        key: str,
    ) -> Any:
        """
        Get a value from a potentially nested key
        """
        keys = key.split("__")
        result = data
        for key in keys:
            result = result.get(key)
            if isinstance(result, list):
                # just turn lists into strings
                return json.dumps(result)
            if result is None:
                return None
        return result

    def _get_subset_from_keys(
        self,
        data: dict[str, Any],
        keys: list[str],
    ) -> dict[str, Any]:
        """
        Create a subset of the dict using the given keys,
        including nested keys
        """
        result = {}
        for key in keys:
            value = self._get_value_from_key(data, key)
            if value is not None:
                keys_split = key.split("__")
                current = result
                for sub_key in keys_split[:-1]:
                    if sub_key not in current:
                        current[sub_key] = {}
                    current = current[sub_key]
                current[keys_split[-1]] = value

        return result

    def _process_data(
        self,
        data: list[dict[str, Any]],
        select_keys: list[str],
    ) -> list[dict[str, Any]]:
        """
        Reduces and flattens data to only include select keys
        """
        reduced_data = [self._get_subset_from_keys(item, select_keys) for item in data]
        return [flatten(item, separator="__") for item in reduced_data]
